"""Exceptions."""


class ImageSearchException(Exception):
    """Parent exception"""


class DoesNotExistException(ImageSearchException):
    """Indicates that a path does not exist."""


class FileNotReadableException(ImageSearchException):
    """Indicates the path is not a file (for source images)."""


class UnsupportedImageException(ImageSearchException):
    """Indicates that the file is not supported for decoding."""


class NotSearchableException(ImageSearchException):
    """Indicates that a path is not a file or directory."""


class ThresholdException(ImageSearchException):
    """Indicates the threshold is invalid."""


class UnknownAlgorithmException(ImageSearchException):
    """Indicates the specified algorithm does not name a known algorithm."""



"""Run the module."""
from __future__ import annotations
import sys
from typing import Optional

from .config import get_config, Config
from .search import search
from .fingerprint import Algorithm
from .exceptions import ImageSearchException


def main(config: Optional[Config] = None) -> None:
    """Run the image matcher."""
    if config is None:
        config = get_config()
    try:
        image_diffs = search(
            ref_path=config.ref_path,
            search_paths=config.search_paths,
            algorithm=Algorithm.from_name(config.algorithm),
            threshold=config.threshold,
            stop_on_first_match=config.stop_on_first_match
        )

        for image_diff in image_diffs:
            print(image_diff.output_line())
    except ImageSearchException as exc:
        sys.stderr.write(f"{exc}\n")
        sys.exit(True)


if __name__ == '__main__':
    main()



"""
Returns the similiarity between a reference image and a set of other images.
"""
__version__ = '0.1.2'

from .fingerprint import Algorithm
from . import exceptions
from .search import search, ImageDiff



"""Functionality around searching paths and determining visual differences."""
from __future__ import annotations

from pathlib import Path
from typing import Generator, List, Optional, Set

import attr

from .exceptions import (
    ThresholdException,
    DoesNotExistException,
    NotSearchableException,
    UnsupportedImageException,
    FileNotReadableException
)
from .fingerprint import Algorithm


@attr.s(frozen=True, auto_attribs=True, kw_only=True, order=True)
class ImageDiff:
    """The difference in hash between this image and a reference image."""
    path: Path = attr.ib(order=False)
    diff: int

    def output_line(self) -> str:
        """Return the output string of this object"""
        return f"{self.diff}\t{self.path}"


@attr.s(frozen=True, auto_attribs=True, kw_only=True, order=False)
class SearchPathFile:
    """
    A file that should be searched.

    path: The Path to the file
    explicit: Whether the file was explicitly asked to be compared (i.e. was provided as a search
        path and not a directory's child). If so, and the file is not a valid image, the we can
        error.
    """
    path: Path
    explicit: bool

    def fs_valid(self) -> bool:
        """
        Returns if the file is valid from an filesystem perspective: it's a file (not a directory,
        etc.)
        """
        return self.path.is_file()


def iterate_paths(search_path: Path) -> Generator[SearchPathFile, None, None]:
    """
    Yield all existing subpaths from a given path.

    This means that if the path is a:
        - file, then yield that file if it is an allowed extension
        - directory, then yield all files under that directory that are allowed extensions
    """
    if search_path.is_file():
        yield SearchPathFile(
            path=search_path,
            explicit=True
        )

    elif search_path.is_dir():
        children_count = 0

        for child_path in search_path.rglob("*"):
            yield SearchPathFile(
                path=child_path,
                explicit=False
            )

        if children_count == 0:
            # warn that there were no valid files in directory
            pass

    else:
        raise NotSearchableException(
            f'Search path "{search_path}" is not a file or directory.'
        )


def check_threshold(threshold: Optional[int]) -> None:
    """Checks the threshold for nonnegativity."""
    if threshold is not None and threshold < 0:
        raise ThresholdException("Threshold must be nonnegative.")


def check_stop_on_first_match(stop_on_first_match: bool, threshold: Optional[int]) -> None:
    """
    Checks that we can stop on the first match, which is only possible if a threshold is specified.
    """
    if threshold is None and stop_on_first_match:
        raise ThresholdException(
            "If stopping on first match, a threshold must be specified."
        )


def search(
        ref_path: Path,
        search_paths: List[Path],
        algorithm: Algorithm,
        stop_on_first_match: bool = False,
        threshold: Optional[int] = None,
) -> Generator[ImageDiff, None, None]:
    """
    Returns a generator of ImageDiff objects for all images in search paths against the ref path
    image.
    """
    check_threshold(threshold)
    check_stop_on_first_match(stop_on_first_match, threshold)

    reference_image_hash = algorithm.hash_path(ref_path)

    seen_images: Set[Path] = set()

    for search_path in search_paths:
        for child_path in iterate_paths(search_path):

            if child_path.path in seen_images and not child_path.explicit:
                continue
            seen_images.add(child_path.path)

            try:
                image_hash = algorithm.hash_path(child_path.path)
            except (
                    UnsupportedImageException,
                    DoesNotExistException,
                    FileNotReadableException
            ):
                if child_path.explicit:
                    raise
                continue

            diff = image_hash - reference_image_hash
            if threshold is None or diff <= threshold:
                yield ImageDiff(
                    path=child_path.path,
                    diff=diff
                )
                if stop_on_first_match:
                    break



"""Functionality related to configuration."""

from __future__ import annotations
import argparse
from typing import Type, List, Optional
from pathlib import Path

import attr

from .fingerprint import Algorithm
from . import __version__


@attr.s(frozen=True, auto_attribs=True, order=False, kw_only=True)
class Config:
    """Holds configuration parameters."""

    ref_path: Path
    search_paths: List[Path]
    algorithm: str
    stop_on_first_match: bool
    threshold: Optional[int]

    @classmethod
    def from_args(cls: Type[Config], args: Optional[List[str]] = None) -> Config:
        """Generate a _Config from command line arguments."""
        parser = argparse.ArgumentParser(
            prog="imagesearch",
            description=(
                "Returns the similiarity between a reference image and a set of other images."
            ),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        parser.add_argument(
            '--version',
            action='version',
            version=__version__
        )

        parser.add_argument(
            'ref_path',
            type=Path,
            help="The reference image to compare among the search paths."
        )

        parser.add_argument(
            'search_path',
            nargs='+',
            type=Path,
            help=(
                "The path(s) of other images to compare against the reference image. This argument "
                "may be supplied multiples times. The path may be either a file (for explicit "
                "comparison) or a directory that will be recursively searched for image files."
            )
        )

        each_algo_help_str = ", ".join(
            f"{algo.algo_name} = {algo.description}" for algo in list(Algorithm)
        )
        parser.add_argument(
            "-a",
            "--algorithm",
            action="store",
            default='ahash',
            choices=Algorithm.algo_names(),
            help=(
                "The algorithm to use to fingerprint the images. See the imagehash library for "
                f"documentation on how each works. {each_algo_help_str}."
            ),
        )

        parser.add_argument(
            '-1',
            '--stop-on-first-match',
            action='store_true',
            help=(
                'Stop searching immediately after the first image with diff <= threshold is found '
                '(instead of continuing to search through all search paths), which also means that '
                'this argument must be used with the threshold argument as well . If multiple '
                'matches do exist, the result is arbitrary. This argument is intended to search '
                'through a large set of images where only a single image is desired.'
            )
        )

        parser.add_argument(
            '-t',
            '--threshold',
            type=int,
            default=None,
            help=(
                'The the maximum difference at which to consider 2 images as a match. Lower '
                'numbers mean closer match. This value is subjective and depends on the hash '
                'algorithm used, so testing without a threshold may be useful to feel for the '
                'values. Omit to use an infinite threshold.'
            )
        )

        parsed_args = parser.parse_args(args=args)

        return cls(
            ref_path=parsed_args.ref_path,
            search_paths=parsed_args.search_path,
            algorithm=parsed_args.algorithm,
            stop_on_first_match=parsed_args.stop_on_first_match,
            threshold=parsed_args.threshold,
        )

def get_config(args: Optional[List[str]] = None) -> Config:
    """Returns the configuration"""
    return Config.from_args(args)



"""Compute visual fingerprints of images."""
from __future__ import annotations

from pathlib import Path
from typing import Callable, Dict, List
from enum import Enum, unique

import imagehash
from PIL import Image, UnidentifiedImageError

from .exceptions import (
    UnsupportedImageException,
    DoesNotExistException,
    FileNotReadableException,
    UnknownAlgorithmException,
)


@unique
class Algorithm(Enum):
    """
    An enumeration of the algorithms provided by imagehash, along with a name and description.
    """
    AHASH = (
        'ahash',
        imagehash.average_hash,
        'Average hashing'
    )
    PHASH = (
        'phash',
        imagehash.phash,
        '2-axis perceptual hashing'
    )
    PHASH_SIMPLE = (
        'phash-simple',
        imagehash.phash_simple,
        '1-axis perceptual hashing'
    )
    DHASH = (
        'dhash',
        imagehash.dhash,
        'Horizontal difference hashing'
    )
    DHASH_VERT = (
        'dhash-vert',
        imagehash.dhash_vertical,
        'Vertical difference hashing'
    )
    WHASH_HAAR = (
        'whash-haar',
        imagehash.whash,
        'Haar wavelet hashing'
    )
    WHASH_DB4 = (
        'whash-db4',
        lambda img: imagehash.whash(img, mode='db4'),
        'Daubechies wavelet hashing'
    )
    COLORHASH = (
        'colorhash',
        imagehash.colorhash,
        'HSV color hashing'
    )

    @classmethod
    def from_name(cls, name: str) -> Algorithm:
        """Returns an algorithm by name."""
        name_map: Dict[str, Algorithm] = {
            algo.algo_name: algo for algo in list(Algorithm)
        }
        if name not in name_map:
            raise UnknownAlgorithmException(f"No algorithm with name {name}")
        return name_map[name]

    @classmethod
    def algo_names(cls) -> List[str]:
        """Return a list of all the algorithm names."""
        return [algo.algo_name for algo in list(Algorithm)]

    def __init__(
            self,
            algo_name: str,
            method: Callable[[Image], imagehash.ImageHash],
            description: str,
    ):
        self.algo_name = algo_name
        self.method = method
        self.description = description

    def hash_path(self, path: Path) -> imagehash.ImageHash:
        """Compute the visual fingerprint of image at path with an algorithm."""
        try:
            image = Image.open(path)
        except UnidentifiedImageError as exc:
            raise UnsupportedImageException(
                f"Path {path} is not a supported image format: {exc}")
        except FileNotFoundError as exc:
            raise DoesNotExistException(
                f"Path {path} could not be found: {exc}")
        except PermissionError as exc:
            raise FileNotReadableException(
                f"Could not open path {path} for reading: {exc}")
        return self.method(image)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imagesearch']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0', 'imagehash>=4.1.0,<5.0.0']

entry_points = \
{'console_scripts': ['imagesearch = imagesearch.__main__:main']}

setup_kwargs = {
    'name': 'imagesearch',
    'version': '0.1.2',
    'description': 'Measure visual similiarity of a reference image to other images.',
    'long_description': '# imagesearch\n\n`imagesearch` measures visual similiarity between a reference image and a set of other\nimages. This can be used to search for a similar image in a large/deep directory structure.\n\n## Installation\n\n    pip install imagesearch\n\n## Examples\n\n- Compare a reference image to all images in a search path:\n\n        > imagesearch needle.jpg haystack\\\n        28      haystack\\0.jpg\n        38      haystack\\1.jpg\n        12      haystack\\2.jpg\n        18      haystack\\3.jpg\n        32      haystack\\4.jpg\n        29      haystack\\5.jpg\n        0       haystack\\6.jpg\n        29      haystack\\7.jpg\n        5       haystack\\8.jpg\n        28      haystack\\9.jpg\n\n    In this example, `haystack\\6.jpg` is most similar.\n\n- Compare against a single image:\n\n        > imagesearch needle.jpg haystack\\1.jpg\n        38       haystack\\1.jpg\n\n- Only return images with similarity less than or equal to 10:\n\n        > imagesearch needle.jpg haystack\\ --threshold 10\n        0       haystack\\6.jpg\n        5       haystack\\8.jpg\n\n- Return the first image found under the threshold (0, in this case) and stop searching immediately:\n\n        > imagesearch needle.jpg haystack\\ -t 0 -1\n        0       haystack\\6.jpg\n\n- Specify a different algorithm:\n\n        > imagesearch needle.jpg haystack\\ --algorithm colorhash\n        ...\n\n- Get more help:\n\n        > imagesearch --help\n        ...\n\n## Visual Similiarity\n\n`imagesearch` returns a nonnegative integer that quantifies the visual similarity between the\nreference image and another image. It does this by creating an image fingerprint and looking at the\ndifference between them.\n\nA critical feature of these fingerprints is that they can be numerically compared (by Hamming Distance).\nImages that are different will have large differences in their fingerprints, and vice versa\n\n**A `0` value indicates the highest level of similarity, or possibly a true match.**\n\nValues should be treated as opaque and relative. It is dependent on the algorithm\nused to create the fingerprints and your subjective criteria for what "similar" is.\n\nThis project uses the\n[imagehash](https://github.com/JohannesBuchner/imagehash) library to produce these fingerprints, and\nmore information about the techniques can be found there.\n\n## Algorithms\n\nAll the fingerprinting algorithms in `imagesearch` come from [imagehash](https://github.com/JohannesBuchner/imagehash). In `imagesearch`, you may specify which algorithm to use by passing the appropriate option value to the `-a` or `--algorithm` flag:\n\n- `ahash`: Average hashing (aHash)\n- `phash`: 2-axis perceptual hashing (pHash)\n- `phash-simple`: 1-axis perceptual hashing (pHash)\n- `dhash`: Horizontal difference hashing (dHash)\n- `dhash-vert`: Vertical difference hashing (dHash)\n- `whash-haar`: Haar wavelet hashing (wHash)\n- `whash-db4`: Daubechies wavelet hashing (wHash)\n- `colorhash`: HSV color hashing (colorhash)\n\nSee\n[this section of the imagehash documentation](https://github.com/JohannesBuchner/imagehash#example-results)\nfor examples of **different methods producing the same fingerprint for different images**. These\nare the analog to cryptographic hash collosions, so it\'s important to understand what kinds of\nscenarios may cause this!\n',
    'author': 'Tim Martin',
    'author_email': 'tim@timmart.in',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/t-mart/imagesearch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
